@@ -20,45 +20,100 @@ import os
 from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
 
-from os_brick.i18n import _LE, _LW
 from os_brick.initiator import linuxscsi
 
 LOG = logging.getLogger(__name__)
 
 
 class LinuxFibreChannel(linuxscsi.LinuxSCSI):
-    def _get_hba_channel_scsi_target(self, hba):
+
+    def has_fc_support(self):
+        FC_HOST_SYSFS_PATH = '/sys/class/fc_host'
+        if os.path.isdir(FC_HOST_SYSFS_PATH):
+            return True
+        else:
+            return False
+
+    def _get_hba_channel_scsi_target(self, hba, conn_props):
         """Try to get the HBA channel and SCSI target for an HBA.
 
         This method only works for Fibre Channel targets that implement a
         single WWNN for all ports, so caller should expect us to return either
-        None or an empty list.
+        explicit channel and targets or wild cards if we cannot determine them.
+
+        The connection properties will need to have "target" values defined in
+        it which are expected to be tuples of (wwpn, lun).
+
+        :returns: List of lists with [c, t, l] entries, the channel and target
+        may be '-' wildcards if unable to determine them.
         """
+        # We want the target's WWPNs, so we use the initiator_target_map if
+        # present for this hba or default to target_wwns if not present.
+        targets = conn_props['targets']
+        if 'initiator_target_map' in conn_props:
+            targets = conn_props['initiator_target_lun_map'].get(
+                hba['port_name'], targets)
+
         # Leave only the number from the host_device field (ie: host6)
         host_device = hba['host_device']
         if host_device and len(host_device) > 4:
             host_device = host_device[4:]
 
         path = '/sys/class/fc_transport/target%s:' % host_device
-        cmd = 'grep %(wwnn)s %(path)s*/node_name' % {'wwnn': hba['node_name'],
-                                                     'path': path}
-        try:
-            out, _err = self._execute(cmd)
-            return [line.split('/')[4].split(':')[1:]
-                    for line in out.split('\n') if line.startswith(path)]
-        except Exception as exc:
-            LOG.error(_LE('Could not get HBA channel and SCSI target ID, '
-                          'reason: %s'), exc)
-            return None
+        ctls = []
+        for wwpn, lun in targets:
+            cmd = 'grep -Gil "%(wwpns)s" %(path)s*/port_name' % {'wwpns': wwpn,
+                                                                 'path': path}
+            try:
+                # We need to run command in shell to expand the * glob
+                out, _err = self._execute(cmd, shell=True)
+                ctls += [line.split('/')[4].split(':')[1:] + [lun]
+                         for line in out.split('\n') if line.startswith(path)]
+            except Exception as exc:
+                LOG.debug('Could not get HBA channel and SCSI target ID, path:'
+                          ' %(path)s*, reason: %(reason)s', {'path': path,
+                                                             'reason': exc})
+                # If we didn't find any paths just give back wildcards for
+                # the channel and target ids.
+                ctls.append(['-', '-', lun])
+        return ctls
 
-    def rescan_hosts(self, hbas, target_lun):
-        for hba in hbas:
-            # Try to get HBA channel and SCSI target to use as filters
-            cts = self._get_hba_channel_scsi_target(hba)
-            # If we couldn't get the channel and target use wildcards
-            if not cts:
-                cts = [('-', '-')]
-            for hba_channel, target_id in cts:
+    def rescan_hosts(self, hbas, connection_properties):
+        LOG.debug('Rescaning HBAs %(hbas)s with connection properties '
+                  '%(conn_props)s', {'hbas': hbas,
+                                     'conn_props': connection_properties})
+        get_ctsl = self._get_hba_channel_scsi_target
+
+        # Use initiator_target_map provided by backend as HBA exclusion map
+        ports = connection_properties.get('initiator_target_lun_map')
+        if ports:
+            hbas = [hba for hba in hbas if hba['port_name'] in ports]
+            LOG.debug('Using initiator target map to exclude HBAs')
+            process = [(hba, get_ctsl(hba, connection_properties))
+                       for hba in hbas]
+
+        # With no target map we'll check if target implements single WWNN for
+        # all ports, if it does we only use HBAs connected (info was found),
+        # otherwise we are forced to blindly scan all HBAs.
+        else:
+            with_info = []
+            no_info = []
+
+            for hba in hbas:
+                ctls = get_ctsl(hba, connection_properties)
+                found_info = True
+                for hba_channel, target_id, target_lun in ctls:
+                    if hba_channel == '-' or target_id == '-':
+                        found_info = False
+                target_list = with_info if found_info else no_info
+                target_list.append((hba, ctls))
+
+            process = with_info or no_info
+            msg = "implements" if with_info else "doesn't implement"
+            LOG.debug('FC target %s single WWNN for all ports.', msg)
+
+        for hba, ctls in process:
+            for hba_channel, target_id, target_lun in ctls:
                 LOG.debug('Scanning host %(host)s (wwnn: %(wwnn)s, c: '
                           '%(channel)s, t: %(target)s, l: %(lun)s)',
                           {'host': hba['host_device'],
@@ -72,6 +127,13 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
 
     def get_fc_hbas(self):
         """Get the Fibre Channel HBA information."""
+
+        if not self.has_fc_support():
+            # there is no FC support in the kernel loaded
+            # so there is no need to even try to run systool
+            LOG.debug("No Fibre Channel support detected on system.")
+            return []
+
         out = None
         try:
             out, _err = self._execute('systool', '-c', 'fc_host', '-v',
@@ -82,13 +144,13 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
             # and systool is not installed
             # 96 = nova.cmd.rootwrap.RC_NOEXECFOUND:
             if exc.exit_code == 96:
-                LOG.warning(_LW("systool is not installed"))
+                LOG.warning("systool is not installed")
             return []
         except OSError as exc:
             # This handles the case where rootwrap is NOT used
             # and systool is not installed
             if exc.errno == errno.ENOENT:
-                LOG.warning(_LW("systool is not installed"))
+                LOG.warning("systool is not installed")
             return []
 
         # No FC HBAs were found
@@ -124,8 +186,6 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
         # Note(walter-boring) modern Linux kernels contain the FC HBA's in /sys
         # and are obtainable via the systool app
         hbas = self.get_fc_hbas()
-        if not hbas:
-            return []
 
         hbas_info = []
         for hba in hbas:
@@ -147,11 +207,10 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
         hbas = self.get_fc_hbas()
 
         wwpns = []
-        if hbas:
-            for hba in hbas:
-                if hba['port_state'] == 'Online':
-                    wwpn = hba['port_name'].replace('0x', '')
-                    wwpns.append(wwpn)
+        for hba in hbas:
+            if hba['port_state'] == 'Online':
+                wwpn = hba['port_name'].replace('0x', '')
+                wwpns.append(wwpn)
 
         return wwpns
 
@@ -161,15 +220,12 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
         # Note(walter-boring) modern Linux kernels contain the FC HBA's in /sys
         # and are obtainable via the systool app
         hbas = self.get_fc_hbas()
-        if not hbas:
-            return []
 
         wwnns = []
-        if hbas:
-            for hba in hbas:
-                if hba['port_state'] == 'Online':
-                    wwnn = hba['node_name'].replace('0x', '')
-                    wwnns.append(wwnn)
+        for hba in hbas:
+            if hba['port_state'] == 'Online':
+                wwnn = hba['node_name'].replace('0x', '')
+                wwnns.append(wwnn)
 
         return wwnns
 
@@ -179,8 +235,6 @@ class LinuxFibreChannelS390X(LinuxFibreChannel):
         """Get Fibre Channel WWNs and device paths from the system, if any."""
 
         hbas = self.get_fc_hbas()
-        if not hbas:
-            return []
 
         hbas_info = []
         for hba in hbas:
@@ -219,8 +273,8 @@ class LinuxFibreChannelS390X(LinuxFibreChannel):
             try:
                 self.echo_scsi_command(zfcp_device_command, "1")
             except putils.ProcessExecutionError as exc:
-                LOG.warning(_LW("port_rescan call for s390 failed exit"
-                                " %(code)s, stderr %(stderr)s"),
+                LOG.warning("port_rescan call for s390 failed exit"
+                            " %(code)s, stderr %(stderr)s",
                             {'code': exc.exit_code, 'stderr': exc.stderr})
 
         zfcp_device_command = ("/sys/bus/ccw/drivers/zfcp/%s/%s/unit_add" %
@@ -229,8 +283,8 @@ class LinuxFibreChannelS390X(LinuxFibreChannel):
         try:
             self.echo_scsi_command(zfcp_device_command, lun)
         except putils.ProcessExecutionError as exc:
-            LOG.warning(_LW("unit_add call for s390 failed exit %(code)s, "
-                            "stderr %(stderr)s"),
+            LOG.warning("unit_add call for s390 failed exit %(code)s, "
+                        "stderr %(stderr)s",
                         {'code': exc.exit_code, 'stderr': exc.stderr})
 
     def deconfigure_scsi_device(self, device_number, target_wwn, lun):
@@ -252,6 +306,6 @@ class LinuxFibreChannelS390X(LinuxFibreChannel):
         try:
             self.echo_scsi_command(zfcp_device_command, lun)
         except putils.ProcessExecutionError as exc:
-            LOG.warning(_LW("unit_remove call for s390 failed exit %(code)s, "
-                            "stderr %(stderr)s"),
+            LOG.warning("unit_remove call for s390 failed exit %(code)s, "
+                        "stderr %(stderr)s",
                         {'code': exc.exit_code, 'stderr': exc.stderr})
