@@ -11,6 +11,8 @@
 #    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 #    License for the specific language governing permissions and limitations
 #    under the License.
+
+import ddt
 import mock
 import os
 import six
@@ -23,6 +25,7 @@ from os_brick.initiator import linuxscsi
 from os_brick.tests.initiator import test_connector
 
 
+@ddt.ddt
 class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
 
     def setUp(self):
@@ -65,13 +68,13 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                  'device_path': hbas[0]['ClassDevicePath']}]
         return info
 
-    def fibrechan_connection(self, volume, location, wwn):
+    def fibrechan_connection(self, volume, location, wwn, lun=1):
         return {'driver_volume_type': 'fibrechan',
                 'data': {
                     'volume_id': volume['id'],
                     'target_portal': location,
                     'target_wwn': wwn,
-                    'target_lun': 1,
+                    'target_lun': lun,
                 }}
 
     @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
@@ -123,8 +126,10 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         location = '10.0.2.15:3260'
         wwn = '1234567890123456'
         connection_info = self.fibrechan_connection(vol, location, wwn)
-        volume_paths = self.connector.get_volume_paths(
-            connection_info['data'])
+        conn_data = self.connector._add_targets_to_connection_properties(
+            connection_info['data']
+        )
+        volume_paths = self.connector.get_volume_paths(conn_data)
 
         expected = ['/dev/disk/by-path/pci-0000:05:00.2'
                     '-fc-0x1234567890123456-lun-1']
@@ -138,7 +143,9 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
     @mock.patch.object(linuxscsi.LinuxSCSI, 'remove_scsi_device')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
-    def test_connect_volume(self, get_device_info_mock,
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
+    def test_connect_volume(self, check_valid_device_mock,
+                            get_device_info_mock,
                             get_scsi_wwn_mock,
                             remove_device_mock,
                             get_fc_hbas_info_mock,
@@ -146,6 +153,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                             realpath_mock,
                             exists_mock,
                             wait_for_rw_mock):
+        check_valid_device_mock.return_value = True
         get_fc_hbas_mock.side_effect = self.fake_get_fc_hbas
         get_fc_hbas_info_mock.side_effect = self.fake_get_fc_hbas_info
 
@@ -164,18 +172,23 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         name = 'volume-00000001'
         vol = {'id': 1, 'name': name}
         # Should work for string, unicode, and list
-        wwns = ['1234567890123456', six.text_type('1234567890123456'),
-                ['1234567890123456', '1234567890123457']]
-        for wwn in wwns:
-            connection_info = self.fibrechan_connection(vol, location, wwn)
+        wwns_luns = [
+            ('1234567890123456', 1),
+            (six.text_type('1234567890123456'), 1),
+            (['1234567890123456', '1234567890123457'], 1),
+            (['1234567890123456', '1234567890123457'], 1),
+        ]
+        for wwn, lun in wwns_luns:
+            connection_info = self.fibrechan_connection(vol, location,
+                                                        wwn, lun)
             dev_info = self.connector.connect_volume(connection_info['data'])
             exp_wwn = wwn[0] if isinstance(wwn, list) else wwn
             dev_str = ('/dev/disk/by-path/pci-0000:05:00.2-fc-0x%s-lun-1' %
                        exp_wwn)
             self.assertEqual(dev_info['type'], 'block')
             self.assertEqual(dev_info['path'], dev_str)
-            self.assertTrue('multipath_id' not in dev_info)
-            self.assertTrue('devices' not in dev_info)
+            self.assertNotIn('multipath_id', dev_info)
+            self.assertNotIn('devices', dev_info)
 
             self.connector.disconnect_volume(connection_info['data'], dev_info)
             expected_commands = []
@@ -183,16 +196,17 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
 
         # Should not work for anything other than string, unicode, and list
         connection_info = self.fibrechan_connection(vol, location, 123)
-        self.assertRaises(exception.NoFibreChannelHostsFound,
+        self.assertRaises(exception.VolumePathsNotFound,
                           self.connector.connect_volume,
                           connection_info['data'])
 
         get_fc_hbas_mock.side_effect = [[]]
         get_fc_hbas_info_mock.side_effect = [[]]
-        self.assertRaises(exception.NoFibreChannelHostsFound,
+        self.assertRaises(exception.VolumePathsNotFound,
                           self.connector.connect_volume,
                           connection_info['data'])
 
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device_path')
     def _test_connect_volume_multipath(self, get_device_info_mock,
                                        get_scsi_wwn_mock,
                                        get_fc_hbas_info_mock,
@@ -202,7 +216,8 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                        wait_for_rw_mock,
                                        find_mp_dev_mock,
                                        access_mode,
-                                       should_wait_for_rw):
+                                       should_wait_for_rw,
+                                       find_mp_device_path_mock):
         self.connector.use_multipath = True
         get_fc_hbas_mock.side_effect = self.fake_get_fc_hbas
         get_fc_hbas_info_mock.side_effect = self.fake_get_fc_hbas_info
@@ -227,7 +242,10 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         vol = {'id': 1, 'name': name}
         initiator_wwn = ['1234567890123456', '1234567890123457']
 
-        find_mp_dev_mock.return_value = '/dev/disk/by-id/dm-uuid-mpath-' + wwn
+        find_mp_device_path_mock.return_value = '/dev/mapper/mpatha'
+        find_mp_dev_mock.return_value = {"device": "dm-3",
+                                         "id": wwn,
+                                         "name": "mpatha"}
 
         connection_info = self.fibrechan_connection(vol, location,
                                                     initiator_wwn)
@@ -240,10 +258,8 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         self.connector.disconnect_volume(connection_info['data'],
                                          devices['devices'][0])
         expected_commands = [
-            'multipath -f ' + wwn,
-            'blockdev --flushbufs /dev/sdb',
+            'multipath -f ' + find_mp_device_path_mock.return_value,
             'tee -a /sys/block/sdb/device/delete',
-            'blockdev --flushbufs /dev/sdc',
             'tee -a /sys/block/sdc/device/delete',
         ]
         self.assertEqual(expected_commands, self.cmds)
@@ -257,7 +273,9 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
     @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
-    def test_connect_volume_multipath_rw(self, get_device_info_mock,
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
+    def test_connect_volume_multipath_rw(self, check_valid_device_mock,
+                                         get_device_info_mock,
                                          get_scsi_wwn_mock,
                                          get_fc_hbas_info_mock,
                                          get_fc_hbas_mock,
@@ -266,6 +284,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                          wait_for_rw_mock,
                                          find_mp_dev_mock):
 
+        check_valid_device_mock.return_value = True
         self._test_connect_volume_multipath(get_device_info_mock,
                                             get_scsi_wwn_mock,
                                             get_fc_hbas_info_mock,
@@ -285,7 +304,9 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
     @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
     def test_connect_volume_multipath_no_access_mode(self,
+                                                     check_valid_device_mock,
                                                      get_device_info_mock,
                                                      get_scsi_wwn_mock,
                                                      get_fc_hbas_info_mock,
@@ -295,6 +316,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                                      wait_for_rw_mock,
                                                      find_mp_dev_mock):
 
+        check_valid_device_mock.return_value = True
         self._test_connect_volume_multipath(get_device_info_mock,
                                             get_scsi_wwn_mock,
                                             get_fc_hbas_info_mock,
@@ -314,7 +336,9 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
     @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
-    def test_connect_volume_multipath_ro(self, get_device_info_mock,
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
+    def test_connect_volume_multipath_ro(self, check_valid_device_mock,
+                                         get_device_info_mock,
                                          get_scsi_wwn_mock,
                                          get_fc_hbas_info_mock,
                                          get_fc_hbas_mock,
@@ -323,6 +347,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                          wait_for_rw_mock,
                                          find_mp_dev_mock):
 
+        check_valid_device_mock.return_value = True
         self._test_connect_volume_multipath(get_device_info_mock,
                                             get_scsi_wwn_mock,
                                             get_fc_hbas_info_mock,
@@ -343,7 +368,9 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
     @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
     @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
     def test_connect_volume_multipath_not_found(self,
+                                                check_valid_device_mock,
                                                 get_device_info_mock,
                                                 get_scsi_wwn_mock,
                                                 get_fc_hbas_info_mock,
@@ -353,6 +380,7 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
                                                 wait_for_rw_mock,
                                                 find_mp_dev_mock,
                                                 discover_mp_dev_mock):
+        check_valid_device_mock.return_value = True
         discover_mp_dev_mock.return_value = ("/dev/disk/by-path/something",
                                              None)
 
@@ -396,3 +424,214 @@ class FibreChannelConnectorTestCase(test_connector.ConnectorTestCase):
         expected = []
         actual = self.connector.get_all_available_volumes()
         self.assertItemsEqual(expected, actual)
+
+    @mock.patch('eventlet.greenthread.sleep', mock.Mock())
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'find_multipath_device')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'wait_for_rw')
+    @mock.patch.object(os.path, 'exists', return_value=True)
+    @mock.patch.object(os.path, 'realpath', return_value='/dev/sdb')
+    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
+    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas_info')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_scsi_wwn')
+    @mock.patch.object(linuxscsi.LinuxSCSI, 'get_device_info')
+    @mock.patch.object(base.BaseLinuxConnector, 'check_valid_device')
+    def test_connect_volume_device_not_valid(self, check_valid_device_mock,
+                                             get_device_info_mock,
+                                             get_scsi_wwn_mock,
+                                             get_fc_hbas_info_mock,
+                                             get_fc_hbas_mock,
+                                             realpath_mock,
+                                             exists_mock,
+                                             wait_for_rw_mock,
+                                             find_mp_dev_mock):
+
+        check_valid_device_mock.return_value = False
+        self.assertRaises(exception.NoFibreChannelVolumeDeviceFound,
+                          self._test_connect_volume_multipath,
+                          get_device_info_mock,
+                          get_scsi_wwn_mock,
+                          get_fc_hbas_info_mock,
+                          get_fc_hbas_mock,
+                          realpath_mock,
+                          exists_mock,
+                          wait_for_rw_mock,
+                          find_mp_dev_mock,
+                          'rw',
+                          True)
+
+    @ddt.data(
+        {
+            "target_info": {
+                "target_lun": 1,
+                "target_wwn": '1234567890123456',
+            },
+            "expected_targets": [
+                ('1234567890123456', 1)
+            ]
+        },
+        {
+            "target_info": {
+                "target_lun": 1,
+                "target_wwn": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 1),
+            ]
+        },
+        {
+            "target_info": {
+                "target_luns": [1, 1],
+                "target_wwn": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 1),
+            ]
+        },
+        {
+            "target_info": {
+                "target_luns": [1, 2],
+                "target_wwn": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 2),
+            ]
+        },
+        {
+            "target_info": {
+                "target_luns": [1, 1],
+                "target_wwns": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 1),
+            ]
+        },
+        {
+            "target_info": {
+                "target_lun": 7,
+                "target_luns": [1, 1],
+                "target_wwn": 'foo',
+                "target_wwns": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 1),
+            ]
+        },
+        # Add the zone map in now
+        {
+            "target_info": {
+                "target_lun": 1,
+                "target_wwn": '1234567890123456',
+            },
+            "expected_targets": [
+                ('1234567890123456', 1)
+            ],
+            "itmap": {
+                '0004567890123456': ['1234567890123456']
+            },
+            "expected_map": {
+                '0004567890123456': [('1234567890123456', 1)]
+            }
+        },
+        {
+            "target_info": {
+                "target_lun": 1,
+                "target_wwn": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 1),
+            ],
+            "itmap": {
+                '0004567890123456': ['1234567890123456',
+                                     '1234567890123457']
+            },
+            "expected_map": {
+                '0004567890123456': [('1234567890123456', 1),
+                                     ('1234567890123457', 1)]
+            }
+        },
+        {
+            "target_info": {
+                "target_luns": [1, 2],
+                "target_wwn": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 2),
+            ],
+            "itmap": {
+                '0004567890123456': ['1234567890123456'],
+                '1004567890123456': ['1234567890123457'],
+            },
+            "expected_map": {
+                '0004567890123456': [('1234567890123456', 1)],
+                '1004567890123456': [('1234567890123457', 2)],
+            }
+        },
+        {
+            "target_info": {
+                "target_luns": [1, 2],
+                "target_wwn": ['1234567890123456', '1234567890123457'],
+            },
+            "expected_targets": [
+                ('1234567890123456', 1),
+                ('1234567890123457', 2),
+            ],
+            "itmap": {
+                '0004567890123456': ['1234567890123456',
+                                     '1234567890123457']
+            },
+            "expected_map": {
+                '0004567890123456': [('1234567890123456', 1),
+                                     ('1234567890123457', 2)]
+            }
+        },
+
+    )
+    @ddt.unpack
+    def test__add_targets_to_connection_properties(self, target_info,
+                                                   expected_targets,
+                                                   itmap=None,
+                                                   expected_map=None):
+        volume = {'id': 'fake_uuid'}
+        wwn = '1234567890123456'
+        conn = self.fibrechan_connection(volume, "10.0.2.15:3260", wwn)
+        conn['data'].update(target_info)
+
+        if itmap:
+            conn['data']['initiator_target_map'] = itmap
+
+        connection_info = self.connector._add_targets_to_connection_properties(
+            conn['data'])
+        self.assertIn('targets', connection_info)
+        self.assertEqual(expected_targets, connection_info['targets'])
+
+        if itmap:
+            self.assertIn('initiator_target_lun_map', connection_info)
+            self.assertEqual(expected_map,
+                             connection_info['initiator_target_lun_map'])
+
+    @ddt.data(('/dev/mapper/<WWN>', True),
+              ('/dev/mapper/mpath0', True),
+              ('/dev/disk/by-path/pci-1-fc-1-lun-1', False))
+    @ddt.unpack
+    @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.remove_scsi_device')
+    @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.requires_flush')
+    @mock.patch('os_brick.initiator.linuxscsi.LinuxSCSI.get_dev_path')
+    def test__remove_devices(self, path_used, was_multipath, get_dev_path_mock,
+                             flush_mock, remove_mock):
+        get_dev_path_mock.return_value = path_used
+        self.connector._remove_devices(mock.sentinel.con_props,
+                                       [{'device': '/dev/sda'}],
+                                       mock.sentinel.device_info)
+        get_dev_path_mock.assert_called_once_with(mock.sentinel.con_props,
+                                                  mock.sentinel.device_info)
+        flush_mock.assert_called_once_with('/dev/sda', path_used,
+                                           was_multipath)
+        remove_mock.assert_called_once_with('/dev/sda',
+                                            flush=flush_mock.return_value)
